
/******************************************************************
 *
 * Name:	Plugin_procs.c
 * 
 * Description: Audio plugin for OpenDVE with background sending 
 *              audio process  
 * 
 * Author:	Tomasz Stachowiak
 *
 * Date:	03/10/96
 *
 * Change Log:
 *
 ********************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#include <Xm/Xm.h>

#include <sys/time.h>
#include <dmedia/audio.h>

/*
 *    stuff for sproc(2) system call to fork a second thread.
 */
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/schedctl.h>

/*
 *    the following define enables 4.3BSD behavior for the kill(3B)
 *    system call, and the include file is necessary for this function.
 */
#define _BSD_SIGNALS
#include <signal.h>

#include "Plugin_procs.h"
#include "AudioPluginProtocol.h"
#include "AL_procs.h"
#include "gsm.h"
#include "adpcm.h"
#include "ncs.h"
#include "Form.h"


/***********************************************************************
 *
 * GLOBAL VARIABLES
 *
 ***********************************************************************/

/* For audio library functions */
extern ALport audio_output_ports[];
extern ALconfig conf[];

/* For NCS */
ncsApplication *ncsapp;      
extern int multicast;

/* For Motif */
/* XtAppContext app;*/
XtWorkProcId workproc;
XtWorkProc recworkproc;

/* Mine */
int current_state;
int code_type=ADPCM;
int send_pid = 0;


/* LOCAL VARIABLES */
/* For GSM codec */
gsm handle;

/* For ADPCM */
struct adpcm_state adpcmstates[MAX_USERS];
struct adpcm_state adpcmsendstate;

Form *form=NULL;

/*
 * GSM specific functions 
 */
gsm_init()
  {
     handle = gsm_create();
  }

gsm_close()
  {
     gsm_destroy(handle);
  }


initADPCMstates()
{
  int i;
  
  adpcmsendstate.valprev = 0;
  adpcmsendstate.index = 0;

  for(i=0; i<MAX_USERS; i++)
    {
      adpcmstates[i].valprev = 0;
      adpcmstates[i].index = 0;
    }
}

/* 
 * Output audio ports functions
 */
init_audio_output_ports()
  {
     int i;

     for(i=0; i<100; i++)
	audio_output_ports[i] = 0;
  }

close_audio_output_ports()
  {
     int i;

     for(i=0; i<100; i++)
       if(audio_output_ports[i] != 0)
	 {
	   ALcloseport(audio_output_ports[i]);
	   ALfreeconfig(conf[i]);
	 }
  }




/**********************************************************************
 *
 * Functions kept for OpenDVE_procs.c compatibility
 *
 *********************************************************************/
void QueryPluginStatus()
{
}

void ProcessClipboardData(char *format, char *data, int length)
{
}



/********************************************************************
 *
 * Frames management
 *
 *******************************************************************/

/************************    GSM    *********************************/

void DecodeGSMFrame(gsm_frame *frames, int part)
{
  char buffer[5][320];
  int i;
  
  if(audio_output_ports[part] == 0)
    {
      OpenOutPort(part);
    }

 for(i=0; i<5; i++)
    {
      gsm_decode(handle, frames[i], (gsm_signal *)buffer[i]);
    }
  
  ALwritesamps(audio_output_ports[part], buffer, 800);

  form->SetVolMeter(audioGetLevel((char *)buffer, 800));
}




void
SendGSMFrame()
{
   gsmData *msg;
   char sig[5][320];
   int size;
   int i;

   msg = (gsmData *)malloc(sizeof(gsmData));
   if (!msg)
   {
     fprintf(stderr, "Message allocation error\n");
     return;
   }
 
   /* Get the audio from mic. */
   GetFrameFromMic((char *)sig);

   form->SetMicMeter(audioGetLevel((char *)sig, 800));

   for(i=0; i<5; i++)
     {
	gsm_encode(handle, (gsm_signal *)sig[i], (gsm_byte *)&msg->frames[i]);
     }

   ncsAppSendData(ncsapp, -1, GSM, msg, sizeof(gsmData));

}

/****************************   ADPCM  ********************************/

void DecodeADPCMFrame(char *inbuf, int part)
{
  char buffer[1600];
  
  /*  schedctl(NDPRI,NDPHIMIN); */

  if(audio_output_ports[part] == 0)
    {
      OpenOutPort(part);
    }
  
  adpcm_decoder(inbuf, (short *)buffer, 800, &adpcmstates[part]);
  
  ALwritesamps(audio_output_ports[part], buffer, 800);

  form->SetVolMeter(audioGetLevel(buffer, 800));
}


void
SendADPCMFrame()
{
   audioPluginMessageStructure *msg;
   char sig[1600];
   int size;

   msg = (audioPluginMessageStructure *)malloc(sizeof(audioPluginMessageStructure));
   if (!msg)
   {
     fprintf(stderr, "Message allocation error\n");
     return;
   }
 
   /* Get the audio from mic. */
   GetFrameFromMic(sig);

   form->SetMicMeter(audioGetLevel(sig, 800));

   adpcm_coder((short *)sig, msg->buf, 800, &adpcmsendstate);

   ncsAppSendData(ncsapp, -1, ADPCM, msg,
		  sizeof(audioPluginMessageStructure));

   free(msg);

}

/*******************************************************************
 *
 * Background process function for sending audio
 *
 *******************************************************************/
void SendFrame()
{
  schedctl(NDPRI,NDPNORMMAX); 

  InitMic();
  while(1)
    {
      if(current_state == START_STATE)
	{
	  switch(code_type)
	    {
	    case GSM:
	      SendGSMFrame();
	      break;
	      
	    case ADPCM:
	      SendADPCMFrame();
	      break;
	    }
	}
      else
	{
	  sleep(1);
	}
    }
  
  /*CloseMic();*/
}


/**************************************************************************
 *
 * Function processing incoming messages
 *
 *************************************************************************/
void 
ProcessPluginMessage(short messageType, void *data, int data_size, 
		     int sender) 
{
   switch (messageType){
      case ADPCM:
	 DecodeADPCMFrame((char *)data, sender);
	 break;

      case GSM:
	 DecodeGSMFrame((gsm_frame *)data, sender);
	 break;
   }
}


void 
CleanupPlugin(void)
{
}


Boolean recWorkProc()
{
  struct timeval timeout;
  fd_set set;
  ncsData *ncsdata;
  int size;
  int sockfd;
  char *data;

  timeout.tv_sec = 0;
  timeout.tv_usec = 10000;

  sockfd = ncsAppGetFD(ncsapp, &set);
  if(select(sockfd+1, &set, 0, 0, &timeout) > 0)
    {
      ncsAppProcessIncoming(ncsapp, &set);
      size = ncsAppGetData(ncsapp, &set, &data);
      if(size > 0)
	{
	  size = size - sizeof(ncsData) + 1;
	  ncsdata = (ncsData *)data;
	  /*printf("Get data from user %d\n", ncsdata->handle);*/
	  ProcessPluginMessage(ncsdata->type, ncsdata->data, size, 
			       ncsdata->handle);
	  free(data);
	}

    }

  return False;
}

void InitApplication(XtAppContext apc)
{
  multicast = 0;
  signal(SIGCLD, SIG_IGN);  
  InitAudio();
  init_audio_output_ports();
  initADPCMstates();
  VolumeChange(50);
  SetMicGain(127);
  current_state = STOP_STATE;
  gsm_init();
  
   /*
    * Start up NCS
    */
  if((ncsapp = (ncsApplication *)ncsAppAttach(AUDIO_PLUGIN_PORT)) == 0)
    {
      fprintf(stderr, "\nApplication already started. Exiting...\n");
      exit(0);
    }

  ncsAppAddCallback(ncsapp, tagAppQuit, QuitApplication);

  send_pid = sproc(SendFrame, PR_SALL);

  if(send_pid == -1)
    {
      fprintf(stderr, "Unable to create sending thread...aborting\n");
      exit(-1);
    }

  recworkproc = (XtWorkProc)XtAppAddWorkProc(apc, recWorkProc, NULL);
}

void QuitApplication()
{
   gsm_close();
   CleanupPlugin();
   close_audio_output_ports();
   if(send_pid != 0)
     {
       kill(send_pid, SIGKILL);
     }
   ncsAppDettach(ncsapp);
   exit(0);
}



/*
 * Buttons pushing callbacks
 */

void send_pushed()
{
  if (current_state == STOP_STATE)
    {
      current_state = START_STATE;
    }
  else
  {
    current_state = STOP_STATE;
  }
}

/*
 * Scale callback
 */
void scale_change(int value)
{
  VolumeChange(value);
}
/*
 * Scale callback
 */

void mic_scale_change(int value)
{
  SetMicGain(255 - (int)((float)value * (float)2.55));
}


void setADPCM()
{
  code_type = ADPCM;
}

void setGSM()
{
  code_type = GSM;
}


short audioGetLevel(char *buffer, int samples)
{
        int i;
        short max=0;

        for(i=0; i<samples; i++)
        {
                if(((short *)buffer)[i] > max)
                        max = ((short *)buffer)[i];
        }
        return max;
}

void registerForm(Form *f)
{
  form = f;
}

Pixel getColor(Widget w, char *name)
{
  Display *dpy;
  Colormap map;
  int scr;
  XColor exact, col;

  dpy = XtDisplay(w);
  scr = DefaultScreen(dpy);
  map = DefaultColormap(dpy, scr);
  XAllocNamedColor(dpy, map, name, &col, &exact);

  return col.pixel;
}

/************************** END Plugin_procs.c **************************/






